    /*///'
 ./%'
,/'    .adAHHHAbn.           e b  i   o    s      |     |    |   |  | ||
%/    adAHY"'"YHAbn
//  .dHY"'     '"YHb.
//  dHY'         'YHb                       ___     ___
//  HH1           1HH                      /  /    /__/     /
//  HH1           1HH  .d888b.            /__/pen /repress /nterface
//  9Hb.         .dHP  887"88b
//  'VHAn.     .aAHA'   "  ;88
//    9HHAbo.adAHAP       4388b        d8888b .d88b 88b 8   8 .d88b 88b
//     '"YTHHHTY"'          781        8b.    8' .7 8 8 8   8 8' .7 8 8
//                     .   ;d81        '888b. 8 .7  8   9. .7 8 .7  8
//                    u8&oo78P'           'V8 8.7   8   '9.7' 8.7   8
//                     '98887'        o888888 '5880 8    'V'  '5880 8
//
//  O3  --  ebios opi-server  ---------------------------- ---- --- -- -
//  Copyright (c) ebios Informationssysteme
//  Written by Marcus Warga
/
                            //////////////////////////////////////////*/
#include <iostream>
#include <sstream>
#include <typeinfo>
#include <new>
using namespace std;
#include "o3included.h"
#include "o3instream.h"
#include "o3outstream.h"
#include "o3image.h"
#include "o3exceptions.h"

  //////////////////////////////////////////////////////////////////////////////////////////////////
 //..public......................................................................................//
//////////////////////////////////////////////////////////////////////////////////////////////////

IncludedImage::IncludedImage()
{
	try {
		opiheader = new OPIHeader();
	}
	catch(bad_alloc) {
		throw ExceptionOutOfMemory(__FILE__, __LINE__);
	}

	status = eLoading;
	textredirectmode = eReturn;
	inopi13block = false;
	inopi20block = false;

	lasttype = eNone;
	lastitem = NULL;
}

IncludedImage::~IncludedImage()
{
	LOb *lob;

	while (!inputstream->eoLOb())
		outputstream->nulldevice(inputstream->read());  //if this image wasn't used then there

	if (status == eNotIncluded) {
		inopi13block = false;
		// inopi20block = false; ?? XXX !
	}

	if (!inopi13block && !inopi20block)
		return;

	while (!inputstream->eof()) {					// is still image data in the stream..
		lob = inputstream->read();

		if (typeid(*lob) == typeid(CommentLOb)) {
			if (CMP(static_cast<CommentLOb*>(lob)->getParm(0), Literals::PostScript::EndObject))
				inopi13block = false;

			if (CMP(static_cast<CommentLOb*>(lob)->getParm(0), Literals::PostScript::EndOPI))
				inopi20block = false;

			outputstream->nulldevice(lob);  // throw comments away

			if (!inopi13block && !inopi20block)
				break;
		} else outputstream->write(lob);
	}
}

BinLOb* IncludedImage::read()
{
	if (!inputstream->eoLOb())
		return static_cast<BinLOb*>(inputstream->read());
	else {
		BinLOb *lob;
		try {
			lob = new BinLOb(0);
		}
		catch(bad_alloc) {
			throw ExceptionOutOfMemory(__FILE__, __LINE__);
		}

		return lob;
	}
}

void IncludedImage::writeImage(OPIVersion version)
{
	bool useopi13 = false;
	bool useopi20 = false;

	if (version == Image::eOPI13)
		useopi13 = true;
	else if (version == Image::eOPI20)
		useopi20 = true;
	else
		useopi13 = useopi20 = true;

	if (!inopi20block)
		useopi20 = false;

	if (useopi13)
		writeOPI13Comments1();

	if (useopi20)
		writeOPI20Comments1();

	writeGraphicsState(Image::eIncluded);

	if (useopi20)
		writeOPI20Comments2();

	writeImageOperands();
	writeImageData();

	if (useopi20)
		writeOPI20Comments3();

	if (useopi13)
		writeOPI13Comments2();
	
	delete this;		// suicide..
}

inline bool IncludedImage::eoimage() const
{
	return inputstream->eoLOb();
}

bool IncludedImage::onComment(const CommentLOb *comment)
{
	     if (CMP(comment->getParm(0), Literals::PostScript::ALDImageFileName))				onALDImageFileName(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageID))					onALDImageID(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDObjectComments))				onALDObjectComments(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageDimensions))			onALDImageDimensions(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageCropRect))				onALDImageCropRect(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageCropFixed))				onALDImageCropFixed(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImagePosition))				onALDImagePosition(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageResolution))			onALDImageResolution(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageColorType))				onALDImageColorType(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageColor))					onALDImageColor(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageTint))					onALDImageTint(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageOverprint))				onALDImageOverprint(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageType))					onALDImageType(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageGrayMap))				onALDImageGrayMap(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ALDImageTransparency))			onALDImageTransparency(*comment);
	else if (comment->getParm(0).find(Literals::PostScript::ALDImageASCIITag,0) != string::npos) onALDImageASCIITag(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::BeginObject))					onBeginObject(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::BeginOPI))						onBeginOPI(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::Distilled))						onDistilled(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ImageFileName))					onImageFileName(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::MainImage))						onMainImage(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::TIFFASCIITag))					onTIFFASCIITag(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ImageDimensions))				onImageDimensions(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ImageCropRect))					onImageCropRect(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ImageOverprint))				onImageOverprint(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::ImageInks))						onImageInks(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::BeginIncludedImage))			onBeginIncludedImage(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::IncludedImageDimensions))		onIncludedImageDimensions(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::IncludedImageType))				onIncludedImageType(*comment);
	else if (CMP(comment->getParm(0), Literals::PostScript::IncludedImageQuality))			onIncludedImageQuality(*comment);
	else if (!comment->getParm(0).find(Literals::PostScript::AddComment))					onAddComment(*comment);

	else if (CMP(comment->getParm(0), Literals::PostScript::BeginData))						onBeginData(*comment);	// start replacing..
	else if (CMP(comment->getParm(0), Literals::PostScript::BeginBinary))					onBeginBinary(*comment);// start replacing..

	else if (CMP(comment->getParm(0), Literals::PostScript::EndObject))						noImageIncluded();
	else if (CMP(comment->getParm(0), Literals::PostScript::EndIncludedImage))				noImageIncluded();

	// %%EndObject, %%EndIncludedImage and %%EndOPI is catched by writeImage() and thrown away.
	// -> noImageIncluded() is only called, when there's no picture (no %%BeginData or %%BeginBinary)
	
	else {
		lasttype = eNone;
		return false;	// ignore comment (_new_ OPI-comments)
	}

	// if the order of the data between some comments is important, use e%COMMENT% / DataRedirectMode
	outputstream->nulldevice(const_cast<CommentLOb*>(comment));
	return true;
}

bool IncludedImage::onText(const TextLOb *text)
{
	switch (textredirectmode) {
		case eReturn:			break;

		case eGraphicsState:	opiheader->headerItem(OPIHeader::HI_graphicsstate)->add(*text);
								outputstream->nulldevice(const_cast<TextLOb*>(text));
								return true;

		case eImageOperands:	opiheader->headerItem(OPIHeader::HI_imageoperands)->add(*text);
								outputstream->nulldevice(const_cast<TextLOb*>(text));
								return true;
	}

	return false;
}

bool IncludedImage::OPI13Included() const
{
	return inopi13block;
}

bool IncludedImage::OPI20Included() const
{
	return inopi20block;
}

IncludedImage::ImageStatus IncludedImage::getStatus() const
{
	return status;
}

  //////////////////////////////////////////////////////////////////////////////////////////////////
 //..private.....................................................................................//
//////////////////////////////////////////////////////////////////////////////////////////////////

bool IncludedImage::CMP(const string& str1, const char *str2)
{
#ifdef _MSC_VER

	if (!stricmp(str1.c_str(), str2))
		return true;
	else
		return false;

#else

	if (!strcasecmp(str1.c_str(), str2)) 
		return true;
	else 
		return false;

#endif
}

bool IncludedImage::CMP(const string& str1, const string& str2)
{
#ifdef _MSC_VER

	if (!stricmp(str1.c_str(), str2.c_str()))
		return true;
	else
		return false;
#else

	if (!strcasecmp(str1.c_str(), str2.c_str()))
		return true;
	else
		return false;

#endif
}

void IncludedImage::onAddComment(const CommentLOb& comment)
{
	if(lasttype = eHeaderItem) {
		if (lastitem != NULL)
			lastitem->add(comment);
	} else if(lasttype = eTIFFTag13)
		(*opiheader->headerItemMap(OPIHeader::HIM_aldimageasciitags))[lasttifftag13].add(comment);
	else if(lasttype = eTIFFTag20)
		(*opiheader->headerItemMap(OPIHeader::HIM_tiffasciitags))[lasttifftag20].add(comment);
	else // == eNone
		outputstream->write(const_cast<CommentLOb*>(&comment));
}

void IncludedImage::onBeginData(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_begindata)->set(comment);
	lasttype = eNone;
	status = eComplete;
}

void IncludedImage::onBeginBinary(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_beginbinary)->set(comment);
	lasttype = eNone;
	status = eComplete;
	textredirectmode = eReturn;
}

void IncludedImage::noImageIncluded()
{
	lasttype = eNone;
	status = eNotIncluded;
	textredirectmode = eReturn;
}

void IncludedImage::onALDImageFileName(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimagefilename)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimagefilename);
}

void IncludedImage::onALDImageID(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimageid)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimageid);
}

void IncludedImage::onALDObjectComments(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldobjectcomments)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldobjectcomments);
}

void IncludedImage::onALDImageDimensions(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimagedimensions)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimagedimensions);
}

void IncludedImage::onALDImageCropRect(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimagecroprect)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimagecroprect);
}

void IncludedImage::onALDImageCropFixed(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimagecropfixed)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimagecropfixed);
}

void IncludedImage::onALDImagePosition(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimageposition)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimageposition);
}

void IncludedImage::onALDImageResolution(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimageresolution)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimageresolution);
}

void IncludedImage::onALDImageColorType(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimagecolortype)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimagecolortype);
}

void IncludedImage::onALDImageColor(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimagecolor)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimagecolor);
}

void IncludedImage::onALDImageTint(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimagetint)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimagetint);
}

void IncludedImage::onALDImageOverprint(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimageoverprint)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimageoverprint);
}

void IncludedImage::onALDImageType(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimagetype)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimagetype);
}

void IncludedImage::onALDImageGrayMap(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimagegraymap)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimagegraymap);
}

void IncludedImage::onALDImageTransparency(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_aldimagetransparency)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_aldimagetransparency);
}

void IncludedImage::onALDImageASCIITag(const CommentLOb& comment)
{
	int a,b,k;
	// %ALDImageAsciiTag<NNN>: <tagtext>
	a = 17;
	b = comment.find(":", a);

	k = atol(comment.substr(a, b-a).c_str());
//	opiheader->headerItemMap(OPIHeader::HIM_aldimageasciitags)->insert(map<int, HeaderItem>::value_type(k, comment));
	(*opiheader->headerItemMap(OPIHeader::HIM_aldimageasciitags))[k] = comment;

	lasttifftag13 = k;
	lasttype = eTIFFTag13;
}

void IncludedImage::onBeginObject(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_beginobject)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_beginobject);

	textredirectmode = eImageOperands ;
	inopi13block = true;
}

void IncludedImage::onBeginOPI(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_beginopi)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_beginopi);
	inopi20block = true;
}

void IncludedImage::onDistilled(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_distilled)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_distilled);
}

void IncludedImage::onImageFileName(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_imagefilename)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_imagefilename);
}

void IncludedImage::onMainImage(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_mainimage)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_mainimage);
}

void IncludedImage::onTIFFASCIITag(const CommentLOb& comment)
{
	int k;
	// %%TIFFASCIITag: <tagnumber> <tagtext>
	k = atol(comment.getParm(1).c_str());
//	opiheader->headerItemMap(OPIHeader::HIM_tiffasciitags)->insert(map<int, HeaderItem>::value_type(k, comment));
	(*opiheader->headerItemMap(OPIHeader::HIM_tiffasciitags))[k] = comment;
		
	lasttifftag20 = k;
	lasttype = eTIFFTag20;
}

void IncludedImage::onImageDimensions(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_imagedimensions)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_imagedimensions);

	textredirectmode = eGraphicsState;
}

void IncludedImage::onImageCropRect(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_imagecroprect)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_imagecroprect);
}

void IncludedImage::onImageOverprint(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_imageoverprint)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_imageoverprint);
}

void IncludedImage::onImageInks(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_imageinks)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_imageinks);
}

void IncludedImage::onBeginIncludedImage(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_beginincludedimage)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_beginincludedimage);
}

void IncludedImage::onIncludedImageDimensions(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_includedimagedimensions)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_includedimagedimensions);
}

void IncludedImage::onIncludedImageType(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_includedimagetype)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_includedimagetype);
}

void IncludedImage::onIncludedImageQuality(const CommentLOb& comment)
{
	opiheader->headerItem(OPIHeader::HI_includedimagequality)->set(comment);
	lasttype = eHeaderItem;
	lastitem = opiheader->headerItem(OPIHeader::HI_includedimagequality);

	textredirectmode = eImageOperands;
}

void IncludedImage::writeImageData()
{
	ostringstream tmp;

	if (!writeIfValid(*opiheader->headerItem(OPIHeader::HI_begindata)))	{
		TextLOb *text;
		try {
			text = new TextLOb;
		}
		catch(bad_alloc) {
			throw ExceptionOutOfMemory(__FILE__, __LINE__);
		}

		tmp << Literals::PostScript::BeginData << " " << opiheader->headerItem(OPIHeader::HI_beginbinary)->getLineDataX() << " Hex Bytes";
		text->assign(tmp.str());
		outputstream->write(text);
	}

	while (!eoimage())
		outputstream->write(read());

	writeText(Literals::NullString);
	writeText(Literals::PostScript::EndData);
}

void IncludedImage::writeImageOperands()
{
	writeIfValid(*opiheader->headerItem(OPIHeader::HI_imageoperands));
}
