    /*///'
 ./%'
,/'    .adAHHHAbn.           e b  i   o    s      |     |    |   |  | ||
%/    adAHY"'"YHAbn
//  .dHY"'     '"YHb.
//  dHY'         'YHb                       ___     ___
//  HH1           1HH                      /  /    /__/     /
//  HH1           1HH  .d888b.            /__/pen /repress /nterface
//  9Hb.         .dHP  887"88b
//  'VHAn.     .aAHA'   "  ;88
//    9HHAbo.adAHAP       4388b        d8888b .d88b 88b 8   8 .d88b 88b
//     '"YTHHHTY"'          781        8b.    8' .7 8 8 8   8 8' .7 8 8
//                     .   ;d81        '888b. 8 .7  8   9. .7 8 .7  8
//                    u8&oo78P'           'V8 8.7   8   '9.7' 8.7   8
//                     '98887'        o888888 '5880 8    'V'  '5880 8
//
//  O3  --  ebios opi-server  ---------------------------- ---- --- -- -
//  Copyright (c) ebios Informationssysteme
//  Written by Marcus Warga
/
                            //////////////////////////////////////////*/
#include <iostream>
#include <typeinfo>
#include <string>
#include <stdlib.h>
using namespace std;
#include "o3instream.h"
#include "o3outstream.h"
#include "o3server.h"
#include "o3image.h"
#include "o3exceptions.h"
#include "o3log.h"
#include "o3path3.h"
#include "o3comment.h"
#include "o3included.h"
#include "o3eps.h"
#include "o3tiff.h"
#include "o3settings.h"
#include "o3path3mapper.h"

namespace OPIServer {
	IncludedImage	*includedimage;
	EPSImage		*epsimage;
	TIFFImage		*tiffimage;
	
	void			imageReady();
	bool			searchImage(Path3 &filename);
	bool			relinkManual(Path3 &filename);
	bool			writeNewImage(const Path3 &filename);
	void			writeOldImage(const Path3 &filename);
	bool			doesfileexist(const Path3 &filename);
}


#ifndef _MSC_VER
void strlwr(char *str)
{
	unsigned a, b;

	b = strlen(str);
	for (a = 0; a < b; ++a) {
		*str = tolower(*str);
		++str;
	}
}
#endif


  //////////////////////////////////////////////////////////////////////////////////////////////////
 //..public......................................................................................//
//////////////////////////////////////////////////////////////////////////////////////////////////

void OPIServer::init(string &configfile)
{
	path3mapper = new Path3Mapper();	// muss VOR settings init'ed werden! - settings traegt die pfade ein
	settings = new Settings(configfile);

	inputstream = new InputStream();
	outputstream = new OutputStream();
	opilog = new Log();
}

void OPIServer::shutdown()
{
	delete opilog;
	delete outputstream;
	delete inputstream;

	delete settings;
	delete path3mapper;
}

void OPIServer::parse(const string& infile, const string& outfile, bool streamin, bool streamout)
{
	LOb *lob;

	if (streamout)
		settings->logtoscreen = false;

	opilog->start(settings->logfilename);

	if (!streamin)
		inputstream->open(infile);
	else
		inputstream->usecinstream();

	if (!streamout)
		outputstream->open(outfile);
	else
		outputstream->usecoutstream();

	opilog->write(Literals::Input+infile+Literals::Output+outfile);

	switch (settings->getOutputFormat()) {
		case Image::eOPI13:		opilog->write(Literals::Format+Literals::OPI13);
								break;
								
		case Image::eOPI20:		opilog->write(Literals::Format+Literals::OPI20);
								break;
		
		case Image::eOPIMixed:	opilog->write(Literals::Format+Literals::OPIMixed);
								break;
	}

	opilog->write(Literals::Parsing);

	try {
		includedimage = new IncludedImage();
	}
	catch(bad_alloc) {
		throw ExceptionOutOfMemory(__FILE__, __LINE__);
	}

	while (!inputstream->eof()) {
		lob = inputstream->read();

		if (typeid(*lob) == typeid(CommentLOb)) {
			if (!includedimage->onComment(static_cast<CommentLOb*>(lob)))
				outputstream->write(lob);

		} else if (!includedimage->onText(static_cast<TextLOb*>(lob)))
			outputstream->write(lob);

		if (includedimage->getStatus() != IncludedImage::eLoading) { // eComplete or eNotIncluded
			imageReady();

			try {
				includedimage = new IncludedImage();
			}
			catch(bad_alloc) {
				throw ExceptionOutOfMemory(__FILE__, __LINE__);
			}
		}
	}

	delete includedimage;

	opilog->write(Literals::Done);
}

  //////////////////////////////////////////////////////////////////////////////////////////////////
 //..private.....................................................................................//
//////////////////////////////////////////////////////////////////////////////////////////////////

void OPIServer::imageReady()
{
	bool replaceimage;


	if (includedimage->OPI13Included() && includedimage->OPI20Included())
		opilog->write(Literals::OPIMixedImage+Literals::EncapsulatedImageFound);

	else if (includedimage->OPI13Included())
		opilog->write(Literals::OPI13Image+Literals::EncapsulatedImageFound);

	else if (includedimage->OPI20Included())
		opilog->write(Literals::OPI20Image+Literals::EncapsulatedImageFound);

	string tmp;
	if (includedimage->header()->headerItem(OPIHeader::HI_aldimagefilename)->isValid())
		tmp = includedimage->header()->headerItem(OPIHeader::HI_aldimagefilename)->getValue();
	else if (includedimage->header()->headerItem(OPIHeader::HI_imagefilename)->isValid())
		tmp = includedimage->header()->headerItem(OPIHeader::HI_imagefilename)->getValue();
	Path3 filename(tmp, true);

	opilog->write(Literals::Filename+filename.getFullPath());

	if (includedimage->header()->headerItem(OPIHeader::HI_aldimageid)->isValid())
		opilog->write(Literals::ID+includedimage->header()->headerItem(OPIHeader::HI_aldimageid)->getValue());
	else if (includedimage->header()->headerItem(OPIHeader::HI_mainimage)->isValid())
		opilog->write(Literals::ID+includedimage->header()->headerItem(OPIHeader::HI_mainimage)->getValue());
	else
		opilog->write(Literals::ID+Literals::NotAvailable);

	if (includedimage->header()->headerItem(OPIHeader::HI_includedimagequality)->isValid()) {
		switch (includedimage->header()->headerItem(OPIHeader::HI_includedimagequality)->getInt(0)) {
			case 0:
			case 1:		opilog->write(Literals::Quality+Literals::LowQuality);
						replaceimage = true;
						break;

			case 2:		opilog->write(Literals::Quality+Literals::MediumQuality);
						replaceimage = true;
						break;
			case 3:
			default:	opilog->write(Literals::Quality+Literals::HighQuality);
						opilog->write(Literals::NoReplacementRequired);
						replaceimage = false;
		}
	} else {
		opilog->write(Literals::Quality+Literals::NoQuality);		// XXX opi13 image - check image dimensions!
		replaceimage = true;
	}


	if (replaceimage || settings->forcereplace) {
		if (settings->forcereplace)
			opilog->write(Literals::ReplacementForced);
		
		if (!searchImage(filename))
			if (!relinkManual(filename)) {
				writeOldImage(filename);
				return;
			}
		if (!writeNewImage(filename))
			writeOldImage(filename);
	} else writeOldImage(filename);
}

bool OPIServer::searchImage(Path3 &filename)	// maps the path.. modifies &filename
{
	opilog->write(Literals::SearchingImage);

	path3mapper->mapFilePath(filename);

	if (doesfileexist(filename)) {
		opilog->write("    "+Literals::Ok+" ("+filename.getFullPath()+")");
		return true;
	}

	opilog->write("    "+Literals::Failed+" ("+filename.getFullPath()+")");
	return false;
}

#ifdef SUCKY_OLD_CRAPHACK
bool OPIServer::searchImage(string& filename)
{
	int a, b, c;
	char loressep, hiressep;
	bool usefilenameonly = false;
	map<string, string>::iterator i, iend;
	char tmp[2048];
	string lotmp;

	//cout << "muaa search... ";

	opilog->write(Literals::SearchingImage);

	//cout << "asdffff\n";

	for (i = settings->paths.begin(); i != settings->paths.end(); i++) {
		//cout << "XX ";

		if (!settings->casesensitive) {
			filename.copy(tmp, filename.length(), 0);
			tmp[filename.length()] = 0;
			strlwr(tmp);
			filename = tmp;
		}


		lotmp = i->first;
		if (lotmp.length())
			loressep = lotmp[lotmp.length()-1];
		else
			usefilenameonly = true;

		//cout << "loressep**" << loressep << "**\n";

/*		if (loressep == '*') {
			loressep = lotmp[lotmp.length()-2];
			lotmp.erase(lotmp.length()-1, 1);	// remove *
			addrelative = true;
			cout << "addrel " << lotmp << " loressep:" << loressep << "\n";
		}*/
		hiressep = i->second[i->second.length()-1];	// always > 0 see o3settings->cpp

		//cout << "hiressep**" << hiressep << "**\n";
		//cout << "filename]]" << filename.c_str() << "))\n";
		//cout << "lotmp]]" << lotmp.c_str() << "[[\n";

		if ((a = filename.find(lotmp)) != string::npos || usefilenameonly) {
			
			if (!usefilenameonly)
				filename.erase(0, a+lotmp.length());	// del lores-path, leave relative stuff
			else {
				unsigned a = filename.rfind('\\', filename.length());
				if (a != string::npos)
					filename.erase(0, a + 1);
				a = filename.rfind(':', filename.length());
				if (a != string::npos)
					filename.erase(0, a + 1);
				a = filename.rfind('/', filename.length());
				if (a != string::npos)
					filename.erase(0, a + 1);
			}

			//cout << "filename **" << filename.c_str() << "**\n";

			// fix up path separators
			c = filename.length();
			for (b = 0; b < c; b++) {
				if (filename[b] == loressep)
					filename[b] = hiressep;
			}

			//cout << "filename **" << filename.c_str() << "**\n";

			filename.insert(0, i->second);				// hires-path

			//cout << "filename **" << filename.c_str() << "**\n";

			if (doesfileexist(filename)) {
				opilog->write("    "+Literals_Ok);

				return true;
			} else
				break;
		} else //cout << "ELSE\n";
	}

	opilog->write("    "+Literals_Failed+" ("+filename+")");
	return false;
}
#endif

bool OPIServer::relinkManual(Path3 &filename)
{
	if (settings->getReplaceMode() == Settings::eAuto)
		return false;

	do {
		opilog->writeLogScreen(Literals::EnterNewImageFilename);
		opilog->writeLogScreen(Literals::Prompt);
		string tmp;
		cin >> tmp;

		if (!tmp.compare(Literals::AbortKey)) {
			opilog->writeLog(Literals::Aborted);
			return false;
		}

		filename.setPath(tmp, true);
		opilog->writeLog(filename.getFullPath());

	} while (!doesfileexist(filename));
	return true;
}

inline bool xCMP(const string& str1, const char *str2)
{
#ifdef _MSC_VER
	if (!stricmp(str1.c_str(), str2))
		return true;
	else 
		return false;
#else
	if (!strcasecmp(str1.c_str(), str2))
		return true;
	else
		return false;
#endif
}

bool OPIServer::writeNewImage(const Path3& filename)
{
	const string &extension = filename.getExtension();

	if (xCMP(extension, "tif") || xCMP(extension, "tiff")) {
		TIFFImage *tiffimage;
		
		try {
			tiffimage = new TIFFImage(filename.getFullPath(), *includedimage);
		}
		catch (ExceptionFOpenFailed) {
//			delete tiffimage;
			return false;
		}
	
		opilog->write(Literals::ReplacingIncludedImage+filename.getFullPath()+Literals::ReplacingIncludedImageWait);

		tiffimage->writeImage(settings->getOutputFormat());
		delete includedimage;
		
		opilog->write("    "+Literals::Done);
		return true;

	} else if (xCMP(extension, "eps")) {
		EPSImage *epsimage;
		
		try {
			epsimage = new EPSImage(filename.getFullPath(), *includedimage);
		}
		catch (ExceptionFOpenFailed) {
//			delete epsimage;
			return false;
		}

		opilog->write(Literals::ReplacingIncludedImage+filename.getFullPath()+Literals::ReplacingIncludedImageWait);

		epsimage->writeImage(settings->getOutputFormat());
		delete includedimage;

		opilog->write("    "+Literals::Done);
		return true;

	} else
		return false;
}

void OPIServer::writeOldImage(const Path3 &filename)
{
	opilog->write(Literals::CopyingIncludedImage);

	if (includedimage->getStatus() != IncludedImage::eNotIncluded) {
		includedimage->writeImage(settings->getOutputFormat());
		opilog->write("    "+Literals::Done);
	} else {
		TIFFImage *tiffimage;
		
		try {
			//tiffimage = new TIFFImage("o3noimage.tif", *includedimage);
                        tiffimage = new TIFFImage(settings->errorimg, *includedimage);
		}
		catch (ExceptionFOpenFailed&) {
//			delete tiffimage;

			//delete includedimage;
			includedimage->writeImage(settings->getOutputFormat());
			opilog->write("    "+Literals::Done);
			return;
		}

		tiffimage->printerrfilename = filename.getFullPath();

		tiffimage->writeImage(settings->getOutputFormat());
		delete includedimage;
		opilog->write("    "+Literals::Done);
	}
}

bool OPIServer::doesfileexist(const Path3 &filename)
{
	FILE *file;

	file = fopen(filename.getFullPath().c_str(), "rb");

	if (file == NULL)
		return false;

	fclose(file);
	return true;
}
