    /*///'
 ./%'
,/'    .adAHHHAbn.           e b  i   o    s      |     |    |   |  | ||
%/    adAHY"'"YHAbn
//  .dHY"'     '"YHb.
//  dHY'         'YHb                       ___     ___
//  HH1           1HH                      /  /    /__/     /
//  HH1           1HH  .d888b.            /__/pen /repress /nterface
//  9Hb.         .dHP  887"88b
//  'VHAn.     .aAHA'   "  ;88
//    9HHAbo.adAHAP       4388b        d8888b .d88b 88b 8   8 .d88b 88b
//     '"YTHHHTY"'          781        8b.    8' .7 8 8 8   8 8' .7 8 8
//                     .   ;d81        '888b. 8 .7  8   9. .7 8 .7  8
//                    u8&oo78P'           'V8 8.7   8   '9.7' 8.7   8
//                     '98887'        o888888 '5880 8    'V'  '5880 8
//
//  O3  --  ebios opi-server  ---------------------------- ---- --- -- -
//  Copyright (c) ebios Informationssysteme
//  Written by Marcus Warga
/
                            //////////////////////////////////////////*/
#include <iostream>
#include <sstream>
#include <new>
using namespace std;
#include "o3image.h"
#include "o3tiff.h"
#include "o3outstream.h"
#include "o3exceptions.h"
#include "o3log.h"
#include "o3settings.h"

char TIFFImage::hexdigit[16] = { '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F' };

  //////////////////////////////////////////////////////////////////////////////////////////////////
 //..public......................................................................................//
//////////////////////////////////////////////////////////////////////////////////////////////////

TIFFImage::TIFFImage(const string& filename, IncludedImage& clonesrc)
{
//	char tmp[300];
	ostringstream tmp;

	opiheader = clonesrc.header()->clone();

	tiff = TIFFOpen(filename.c_str(), "r");
	if (tiff == NULL) { 
		error = true;
		throw ExceptionFOpenFailed(filename, __FILE__, __LINE__);
		return;
	} else
		error = false;

	TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &imagewidth);
	TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &imageheight);
	TIFFGetField(tiff, TIFFTAG_PLANARCONFIG, &planarconfig);
	TIFFGetField(tiff, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);

	scanlinesize = TIFFScanlineSize(tiff);

	try {
		scanline = new BinLOb(scanlinesize);
	}
	catch(bad_alloc) {
		throw ExceptionOutOfMemory(__FILE__, __LINE__);
	}
	scanline->size(scanlinesize);

	tiffdatasize = imagewidth * imageheight * samplesperpixel;
	if (settings->dataformat == Settings::eHex)
		tiffdatasize <<= 1;

	readrow = 0;

	if (planarconfig != PLANARCONFIG_CONTIG) {
		opilog->write("error - not yet supported: tiff file is not contiguous... "); //+__FILE__+" "+__LINE__);
		
		throw ExceptionFOpenFailed(filename, __FILE__, __LINE__);
	}

	tmp << Literals::PostScript::IncludedImageDimensions << " " << imagewidth << " " << imageheight;
	opiheader->headerItem(OPIHeader::HI_includedimagedimensions)->set(tmp.str());

	tmp.str(Literals::NullString);
	tmp << Literals::PostScript::IncludedImageQuality << " 3";
	opiheader->headerItem(OPIHeader::HI_includedimagequality)->set(tmp.str());

	tmp.str(Literals::NullString);
	tmp << Literals::PostScript::IncludedImageType << " 3 8";
	opiheader->headerItem(OPIHeader::HI_includedimagetype)->set(tmp.str());

	tmp.str(Literals::NullString);
	tmp << Literals::PostScript::ALDImageType << " 3 8";
	opiheader->headerItem(OPIHeader::HI_aldimagetype)->set(tmp.str());

	if (!opiheader->headerItem(OPIHeader::HI_aldimageposition)->isValid()) {
		tmp.str(Literals::NullString);
		tmp << Literals::PostScript::ALDImagePosition << " 0 0 0 " << imageheight << " " << imagewidth << " " << imageheight << " " << imagewidth << " 0";
		opiheader->headerItem(OPIHeader::HI_aldimageposition)->set(tmp.str());
	}
}

TIFFImage::~TIFFImage()
{
	if (!error)
		delete scanline;

	delete opiheader;
}

BinLOb* TIFFImage::read()
{
	if (settings->dataformat == Settings::eHex) {
		BinLOb *hexline;
		unsigned long x;
		char *dst, *src;

		try {
			hexline = new BinLOb(scanlinesize * 2);
		}
		catch(bad_alloc) {
			throw ExceptionOutOfMemory(__FILE__, __LINE__);
		}

		hexline->size(scanlinesize * 2);

		if (TIFFReadScanline(tiff, scanline->rwData(), readrow, 0) == -1)
			throw ExceptionReadScanlineFailed(__FILE__, __LINE__);

		dst = hexline->rwData();
		src = scanline->rwData();

		for (x = 0; x < scanlinesize; x++) {
			*dst = hexdigit[(*src >> 4) & 15];
			dst++;
			*dst = hexdigit[*src & 15];
			dst++;
			src++;
		}

		readrow++;
		return hexline;

	} else { // eBin
		BinLOb *dataline;

		try {
			dataline = new BinLOb(scanlinesize);
		}
		catch(bad_alloc) {
			throw ExceptionOutOfMemory(__FILE__, __LINE__);
		}

		dataline->size(scanlinesize);

		if (TIFFReadScanline(tiff, dataline->rwData(), readrow, 0) == -1)
			throw ExceptionReadScanlineFailed(__FILE__, __LINE__);

		readrow++;
		return dataline;
	}
}

void TIFFImage::writeImage(OPIVersion version)
{
	bool useopi13 = false;
	bool useopi20 = false;

	if (version == Image::eOPI13)
		useopi13 = true;
	else if (version == Image::eOPI20)
		useopi20 = true;
	else if (version == Image::eOPIMixed)
		useopi13 = useopi20 = true;

	if (useopi13)
		writeOPI13Comments1();

	if (useopi20)
		writeOPI20Comments1();

	writeGraphicsState(Image::eTIFF);

	if (useopi20)
		writeOPI20Comments2();

	writeImageOperands();
	writeImageData();

	if (useopi20)
		writeOPI20Comments3();

	if (useopi13)
		writeOPI13Comments2();
	
	delete this;		// suicide..
}

inline bool TIFFImage::eoimage() const
{
	if (readrow >= imageheight)
		return true;
	else
		return false;
}

  //////////////////////////////////////////////////////////////////////////////////////////////////
 //..private.....................................................................................//
//////////////////////////////////////////////////////////////////////////////////////////////////

void TIFFImage::writeImageData()
{
	ostringstream tmp;

	if (settings->dataformat == Settings::eHex)
		tmp << Literals::PostScript::BeginData << " " << tiffdatasize+12 << " Hex Bytes";
	else
		tmp << Literals::PostScript::BeginData << " " << tiffdatasize+12 << " Binary Bytes";

	writeText(tmp.str());
	writeText(Literals::PostScript::Colorimage);

	while (!eoimage())
		outputstream->write(read());

	writeText(Literals::NullString);
	writeText(Literals::PostScript::EndData);
}

void TIFFImage::writeImageOperands()
{
//	char tmp[300];
	ostringstream tmp;
	int i;

//	sprintf(tmp, Literals::PostScript::TIFFCodeLines1[0].c_str(), imagewidth*3);
	tmp << "/EBIOS_picstr " << imagewidth*samplesperpixel << " string def";
	writeText(tmp.str());

//	sprintf(tmp, Literals::PostScript::TIFFCodeLines1[1].c_str(), imagewidth, imageheight);
	tmp.str(Literals::NullString);
	tmp << imagewidth << " " << imageheight << " 8";
	writeText(tmp.str());

//	sprintf(tmp, Literals::PostScript::TIFFCodeLines1[2].c_str(), imagewidth, imageheight, imageheight);
	tmp.str(Literals::NullString);
	tmp << "[" << imagewidth << " 0 0 -" << imageheight << " 0 " << imageheight << "]";
	writeText(tmp.str());

	if (settings->dataformat == Settings::eHex) {
		for (i = 0;; i++) {
			if (!Literals::PostScript::TIFFCodeLines2Hex[i].length()) break;
			writeText(Literals::PostScript::TIFFCodeLines2Hex[i]);
		}
	} else {
		for (i = 0;; i++) {
			if (!Literals::PostScript::TIFFCodeLines2Bin[i].length()) break;
			writeText(Literals::PostScript::TIFFCodeLines2Bin[i]);
		}
	}

	tmp.str(Literals::NullString);
	tmp << samplesperpixel;
	writeText(tmp.str());
}
